/**
 * wackyd -- An LED notification daemon
 *
 * Copyright (C) 2000 Jeffrey Fulmer <jdfulmer@armstrong.com>
 * This file is part of wackyd
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */  
#define  INTERN  1 

#include <setup.h>
#include <getopt.h>
#include <pattern.h>
#include <stdlib.h>
#include <sock.h>
#include <util.h>
#include <list.h>
#include <protocol.h>

/**
 * long options, std options struct
 */
static struct option long_options[] =
{
  { "version",    no_argument,       NULL, 'v' },
  { "help",       no_argument,       NULL, 'h' },
  { "config",     no_argument,       NULL, 'C' },
  { "classic",    no_argument,       NULL, 'c' },
  { "blink",      no_argument,       NULL, 'b' },
  { "speed",      required_argument, NULL, 's' },
  { "port",       required_argument, NULL, 'p' },
  { "kill",       no_argument,       NULL, 'k' }
};

/**
 * display_version   displays the version number and exits
 *                   on boolean false.
 */
void
display_version( int i )
{
  /**
   * version_string is defined in version.c
   * adding it to a separate file allows us
   * to parse it in configure.
   */
  #ifdef DEBUG
  printf( "%s %s: debugging enabled\n", "wackyd", version_string );
  #else
  printf( "%s %s\n", "wackyd", version_string );
  #endif
  /**
   * if TRUE, exit
   */
  if( i == 1 ){ exit( 0 ); }
}  /* end of display version */


/**
 * display_help      displays the help section to STDOUT and exits
 */
void
display_help()
{
  /**
   * call display_version, but do not exit
   */
  display_version( FALSE );
  printf("Usage: %s [options] hostaddr\n", "wackyd" );
  printf("Options:\n"                    );
  puts("  -v, --version         VERSION, prints version number to screen.");
  puts("  -h, --help            HELP, prints this section.");
  puts("  -C, --config          CONFIGURATION, show the current configuration."); 
  puts("  -c, --classic         CLASSIC, start wackyd in classic mode in which it");
  puts("                        does NOT listen for requests; it just blinks.");
  puts("  -p, --port=N          PORT, change default port 9193 to N" );
  puts("  -s, --speed=N         SPEED, set the default blinking speed in ");
  puts("                        tenths of a second. 1 is fast, 10 is slower.");
  puts("  -k, --kill            KILL, shutdown a running wackyd daemon");
  /**
   * our work is done, exit nicely
   */
  exit(0);
}

/**
 * parse_cmdline      parses command line arguments and assigns
 *                    values to run time variables.
 */
void
parse_cmdline( int argc, char *argv[] )
{
  int c = 0;
  int nhosts;

  while ((c = getopt_long (argc, argv, "vhCcks:", long_options, (int *)0)) != EOF){
  switch( c ){
      case 'v':
        display_version( TRUE );
        break;
      case 'h':
        display_help();
        exit( 0 );
      case 'C':
        my.config = TRUE;
        break;
      case 'c':
        my.classic = TRUE;
        break;
      case 'k':
        die_hard( TRUE, ALL );
        exit( 0 );
      case 's':
        my.speed = atoi( optarg );
        break; 
    } /** end of switch( c )           **/
  }   /** end of while c = getopt_long **/
  nhosts      = argc - optind;
  if( !nhosts )
    my.hostname = (strlen(my.hostname)>2)?my.hostname:"*";
  else
    my.hostname = argv[argc-1];
} /** end of parse_cmdline **/

void
clean_up( void )
{
  my.done = TRUE; 
}

void
signal_handler()
{
  int ret;
  ret = restore_lights( my.console, my.state ); 
  close( my.sock ); 

  exit( ret );
}

int
main( int argc, char *argv[] )
{
  long int  state;
  int       result;
  int       csock;
  int       childpid;
  struct    sigaction action; 
  socklen_t addrlen,  len;
  struct    sockaddr_in clnt;
  uid_t     uid;

  init_config(); 
  parse_cmdline(argc, argv);
  if( my.config ){
    show_config( TRUE );    
  }
  init_console( &my.console, &my.state );

  if( file_exists( PIDFILE )) die_hard( FALSE, ALL );

  joe_openlog();

  memset( &action, 0, sizeof( action ));
  action.sa_handler = signal_handler;
  if( sigaction( SIGHUP, &action, NULL ))
    joe_error( "sigaction" );
  if( sigaction( SIGINT, &action, NULL))
    joe_error( "sigaction" );
  if( sigaction( SIGTERM, &action, NULL))
    joe_error( "sigaction" );  
  
  switch( fork() ){
  case -1:                          /* unable to fork  */
    joe_fatal( "unable to fork" );
  case 0:                           /* we are a daemon */
    nice( 10 );

    if(( update_pidfile((long)getpid(),PARENT)) < 0 ){
      joe_error( "Couldn't open pid file!\nYou will have to kill wackyd manually..." );  
    }
  
    /**
     * classic mode  
     */
    if( my.classic ){ blink( my.console, 300000 ); while( TRUE ); } 

    if(( childpid = fork()) == 0 ){
      if(( update_pidfile((long)getpid(),CHILD)) < 0 ){
        joe_error( "Couldn't open pid file!\nYou will have to kill wackyd manually..." );
      } 
      blink( my.console, 300000 );
    }

    fclose( stdin  );
    /*fclose( stderr );*/
    signal( SIGCHLD, SIG_IGN );

    my.hostname  = (my.hostname == NULL)?  "*" :my.hostname;
    my.portname  = (my.portname == NULL)?"9193":my.portname;   

    if(( my.sock = WACKYsocket( my.hostname, my.portname, &addrlen )) < 0 ){
      exit( 1 ); /* error handling is in WACKYsocket */ 
    } 
    
    
    for( ;; ){
      char addr[256];
      char name[256];
      len    = addrlen;
      if(( csock = WACKYaccept( my.sock, &clnt, &len )) < 0 ){
        joe_error( "WACKYaccept" );
        exit( 1 );
      } 
      getpeerinfo( csock, addr, name, sizeof( name ));

      die_hard( FALSE, CHILD );
      if(( childpid = fork()) == 0 ){   
        if(( update_pidfile((long)getpid(),CHILD)) < 0 ){
          joe_error( "Couldn't open pid file!\nYou will have to kill wackyd manually..." );
        }
        if( host_match( name, addr ) < 1 ){
          FILE *out = fdopen( csock, "w" );
          fprintf( out, "403 access forbidden.\n" );
          joe_error( "%s ==> %s access denied", addr, name ); 
          fclose( out );
          blink( my.console, 300000 );
        }
        else{
          joe_error( "%s ==> %s authorized", addr, name ); 
          protocol_handler( csock );  
        } 
        exit( 0 );
      } 
      close( csock );
    }
    joe_closelog();
  }
  exit( 0 );

}
